#include <algorithm>

enum
{
    Up      = 2,
    Down    = 4,
    Shutter = 8,
    Right   = 1,
    End     = 1,
    Last    = 0
};

// Routine $C39C: load the page of the level

// In each page, these settings are constant:
static const unsigned Addr_Blocks   = 0x0000; // Max blocks: 192 (shared)
static const unsigned Addr_BlockPals= 0x0300; // Max blocks: 192 (shared)
static const unsigned Addr_Rooms    = 0x03C0; // Max rooms: 33   (shared)
static const unsigned Addr_Order    = 0x0C00; // Max screens: 48 (shared)
static const unsigned Addr_Pointers = 0x0C30; // Max rooms: 32.  (shared)
static const unsigned Addr_Layout   = 0x0C70; // Max screens: 48. (shared)
static const unsigned Addr_Palette1 = 0x0CA0;
static const unsigned Addr_Palette2 = 0x0CD0;
static const unsigned Addr_Tiles1   = 0x0D00; // Map1 ?table (number, followed by 2 bytes each)
static const unsigned Addr_Tiles2   = 0x0D20; // Map2 ?table
static const unsigned Addr_Tiles3a  = 0x0D40; // Something common to both
static const unsigned Addr_Tiles3b  = 0x0D80; // Something common to both
static const unsigned Addr_Actives1 = 0x0E00;
static const unsigned Addr_Actives2 = 0x0EA0;
static const unsigned Addr_ActivesEnd=0x0F00; // begin gfx sprites
static const unsigned Addr_Tiles3c   =0x0F80; // Something common to both, also F70.
  // E00: 1 byte: number of objects (map1) Including terminator
  //      6 bytes per object. Terminator = 2nd byte is $FF
  //      Max objects: 29
  //  0: type
  //       2=gutsblock, 3=door_right, 1=door_left, 0=appearing block
  //  1: screen #
  //  2: x1 (in)
  //  3: y1 (in)
  //  4: x2 (out)
  //  5: y2 (out)
  // EB0: 1 byte: number of objects (map2) Including terminator
  //      6 bytes per object. Terminator = 2nd byte is $FF
  //      Max objects: 13
  // F00: gfx

static const unsigned Addr_EnemyPointers = 0x1A452; //C,I,B,F,E,G,1,2,3,4,? point to A468+.
static /***/ unsigned Addr_EnemyData     = 0x1A468; // room for 419 enemies (total).
static const unsigned Addr_EnemyEnd      = 0x1AB00;

/* Address of the 10-byte table which lists the offsets of
 * each level's order table, relative to the page's Addr_Order.
 */
static const unsigned Addr_OrderOffsetTable = 0x1C2EB;

  // Enemies is list of 4-byte records, terminated by byte $FF.
  //  0: screen #
  //  1: x (center)
  //  2: y (center)
  //  3: type

/* Common ascii characters in all levels */
static const char Ascii_Guts     = 'G';
static const char Ascii_Spike    = '^';
static const char Ascii_Ladder   = '|';
static const char Ascii_WilyDoor = 'D';

static const char Ascii_ObjShutterIn  = 'X'; // input shutter
static const char Ascii_ObjShutterOut = 'S'; // output shutter
// pneumatic  blocks are represented by hex characters 0..F
static const char Ascii_ObjFire    = 'I'; // flying fire

// A level may contain 48 rooms, but when two levels are paired, the total amount must <= 48.
static const unsigned MaxRooms   = 48;
// There is room for 13 actives at max per level.
static const unsigned MaxActives = 32;
// THere is room for 419 enemies at max per level.
static const unsigned MaxEnemies = 256;

struct Level
{
    unsigned LevelNumber;
    const char* Name;
    
    Screen rooms[MaxRooms];
    Active actives[MaxActives];
    Enemy  enemies[MaxEnemies];

    unsigned char Palette[16*3];
    
    /* Each level has different representations
     * in text form.
     */
    void (*CreateAscii)(Screen& screen);
    void (*CreateChar)(Screen& screen);
    
    void WritePalette(unsigned addr) const
    {
        for(unsigned a=0; a<16*3; ++a)
            ROM[addr+a] = Palette[a];
    }
    void WriteActives(unsigned first_screen, unsigned addr, unsigned end_addr) const
    {
        unsigned n=0;
        for(unsigned a=0; a<MaxActives; ++a)
        {
            if(actives[a].screen == 255) continue;
            unsigned begin = addr + 1 + n*6;
            if(begin+6 > end_addr)
            {
                fprintf(stderr, "Error: Too many actives!\n");
            }
            ROM[begin+0] = actives[a].type;
            ROM[begin+1] = actives[a].screen - first_screen;
            ROM[begin+2] = actives[a].x1;
            ROM[begin+3] = actives[a].y1;
            ROM[begin+4] = actives[a].x2;
            ROM[begin+5] = actives[a].y2;
            ++n;
        }
        ROM[addr] = n + 1;
        for(unsigned a = addr + 1 + n*6; a < end_addr; ++a)
            ROM[a] = 0xFF;
    }
    void WriteEnemies(unsigned first_screen,
                      unsigned ptraddr, unsigned& addr, unsigned end_addr) const
    {
        ROM[ptraddr+0] = ((addr|0x8000) & 0x00FF);
        ROM[ptraddr+1] = ((addr|0x8000) & 0xFF00) >> 8;
        for(unsigned a=0; a<MaxEnemies; ++a)
        {
            if(enemies[a].screen == 255)
            {
                if(addr+1 > end_addr)
                {
                    fprintf(stderr, "Error: Too many enemies (can't write end byte)!\n");
                }
                ROM[addr+0] = 0xFF;
                addr += 1;
                break;
            }
            else
            {
                if(addr+4 > end_addr)
                {
                    fprintf(stderr, "Error: Too many enemies (%X,%X)!\n", addr,end_addr);
                }
                ROM[addr+0] = enemies[a].screen - first_screen;
                ROM[addr+1] = enemies[a].x;
                ROM[addr+2] = enemies[a].y;
                ROM[addr+3] = enemies[a].type;
                addr += 4;
            }
        }
    }
    void ReadPalette(unsigned addr)
    {
        for(unsigned a=0; a<16*3; ++a)
            Palette[a] = ROM[addr+a];
    }
    void ReadActives(unsigned first_screen, unsigned addr)
    {
        unsigned n = ROM[addr++];
        if(n > MaxActives)n=MaxActives;
        for(unsigned a=0; a<MaxActives; ++a) actives[a].Delete();
        for(unsigned a=0; a<n; ++a)
        {
            if(ROM[addr+a*6+1] == 255) break;
            //fprintf(stderr, "Reading active from %X\n", addr);
            actives[a].type   = ROM[addr+a*6+0];
            actives[a].screen = ROM[addr+a*6+1] - first_screen;
            actives[a].x1     = ROM[addr+a*6+2];
            actives[a].y1     = ROM[addr+a*6+3];
            actives[a].x2     = ROM[addr+a*6+4];
            actives[a].y2     = ROM[addr+a*6+5];
        }
    }
    void ReadEnemies(unsigned first_screen, unsigned ptraddr)
    {
        unsigned data_addr = ROM[ptraddr+0] | (ROM[ptraddr+1]<<8);
        data_addr = (ptraddr & 0x1C000) + (data_addr & 0x3FFF);
        //fprintf(stderr, "Enemies %04X -> %06X\n", ptraddr, data_addr);
        for(unsigned a=0; a<MaxEnemies; ++a) enemies[a].Delete();
        for(unsigned a=0; a<MaxEnemies; ++a)
        {
            if(ROM[data_addr+a*4] == 255) break;
            enemies[a].screen = ROM[data_addr+a*4+0] - first_screen;
            enemies[a].x      = ROM[data_addr+a*4+1];
            enemies[a].y      = ROM[data_addr+a*4+2];
            enemies[a].type   = ROM[data_addr+a*4+3];
        }
    }
    
    void SortActives()
    {
        std::sort(actives, actives+MaxActives);
    }
    void SortEnemies()
    {
        std::sort(enemies, enemies+MaxEnemies);
    }
    
    void AnalyzeObjects()
    {
        /* To be called after everything has been read from the ROM
         * and the ascii map has been generated, before dumping them
         * to output.
         */
        
        /* Delete actives that can be expressed on the map. */
        DeleteGutsBlockActives();
        
        /* Move the shutters from the active list to the object list. */
        MakeAsciiActives();
        
        /* Move movable enemies to the ascii map. */
        MakeAsciiEnemies();
        
        /* Last, sort the arrays. */
        SortActives();
        SortEnemies();
    }
    
    void DeleteGutsBlockActives()
    {
        /* Gutsblocks. */
        for(unsigned s=0; s<MaxRooms; ++s)
        {
            Screen& sc = rooms[s];
            
            for(unsigned y=0; y<16; ++y)
                for(unsigned x=0; x<16; ++x)
                  if(sc.ascii[y*32+x] == Ascii_Guts)
                      for(unsigned a=0; a<MaxActives; ++a)
                      {
                          Active& act = actives[a];
                          if(act.screen == s
                          && act.type == 2
                          && act.x1 == x*16
                          && act.y1 == y*16
                          && act.x2 == (((x+2)*16) & 0xFF)
                          && act.y2 == (((y+2)*16) & 0xFF))
                          {
                              act.Delete();
                          }
                      }
        }
    }



/*
69
83
88
8B
A0
BA
D8




*/

    void MakeAsciiActives()
    {
        unsigned buzzcounter[MaxRooms] = {0};
        
        // Buzzblocks are ordered last-to-first.
        // Thus we have to walk the structure backwards.
        for(unsigned a=MaxActives; a-- > 0; )
        {
            Active& act = actives[a];
            if(act.screen == 255) continue;
            if(act.type == 1 && act.y2 <= act.y1)
            {
                // flying fire
                unsigned x1 = act.x1; x1 /= 16;
                unsigned x2 = act.x2; x2 /= 16;
                unsigned y = act.y1; y /= 16;
                for(unsigned x=x1; x<x2; ++x)
                    rooms[act.screen].ascii[y*32+16+x] = Ascii_ObjFire;
                
                act.Delete();
            }
            if(act.type == 1 && act.y2 > act.y1)
            {
                // leftside shutter
                unsigned x1 = act.x1; x1 /= 16;
                unsigned y1 = act.y1; y1 /= 16;
                unsigned x2 = act.x2; if(x2==0) x2=0x100;  x2 /= 16;
                unsigned y2 = act.y2; if(y2==0) y2=0x100;  y2 /= 16;
                
                for(unsigned y=y1; y<y2; ++y)
                    for(unsigned x=x1; x<x2; ++x)
                        rooms[act.screen].ascii[y*32+16+x] = Ascii_ObjShutterIn;
                
                act.Delete();
            }
            if(act.type == 3)
            {
                // rightside shutter
                unsigned x1 = act.x1; x1 /= 16;
                unsigned y1 = act.y1; y1 /= 16;
                unsigned x2 = act.x2; if(x2==0) x2=0x100;  x2 /= 16;
                unsigned y2 = act.y2; if(y2==0) y2=0x100;  y2 /= 16;
                
                for(unsigned y=y1; y<y2; ++y)
                    for(unsigned x=x1; x<x2; ++x)
                        rooms[act.screen].ascii[y*32+16+x] = Ascii_ObjShutterOut;
                
                act.Delete();
            }

            if(act.type == 0 && false)
            {
                // buzz (appearing and disappearing blocks)
                unsigned x1 = act.x1; x1 /= 16;
                unsigned y1 = act.y1; y1 /= 16;
                unsigned x2 = act.x2; if(x2==0) x2=0x100;  x2 /= 16;
                unsigned y2 = act.y2; if(y2==0) y2=0x100;  y2 /= 16;
                
                for(unsigned y=y1; y<y2; ++y)
                    for(unsigned x=x1; x<x2; ++x)
                    {
                        unsigned& count = buzzcounter[act.screen];
                        if(count > 15)
                        {
                            fprintf(stderr, "Too many buzzblocks in one screen!\n");
                            break;
                        }
                        rooms[act.screen].ascii[y*32+16+x] = "0123456789ABCDEF"[count++];
                    }
                
                act.Delete();
            }
        }
    }
    
    void MakeAsciiEnemies()
    {
        for(unsigned a=0; a<MaxEnemies; ++a)
        {
            static const char objects[] =
            {
              //  0
              0,0,0,0,
              //  4
              0,0,0,0,
              //  8
              0,0,0,0,
              // 10
              0,0,0,0,
              // 14
              0,0,0,0,
              // 18
              0,0,0,0,
              // 1C
              0,0,0,0,
              // 20
              0,0,0,0,
              // 24
              0,0,0,0,
              // 28
              0,0,0,0,
              // 2C
              0,0,0,0,
              // 30
              0,0,0,0,
              // 34
              0,0,0,0,
              // 38
              0,0,0,0,
              // 3C
              'b','p','e','P',
              // 40
              'E','U','M',0,
              // 44
              '*'
            };
            Enemy& e = enemies[a];
            if(e.screen == 255) continue;
            
            char ch = 0;
            if(e.type < sizeof(objects))
                ch = objects[e.type];
            
            if(ch != 0)
            {
                unsigned x = e.x / 16;
                unsigned y = e.y / 16;
                
                rooms[e.screen].ascii[y*32+16+x] = ch;
                e.Delete();
            }
        }
    }
};
struct LevelPairData
{
    Level& level1;
    Level& level2;
    
    unsigned BeginAddr;
};
